@@ -13,17 +13,16 @@ from xonsh.dirstack import cd, pushd, popd, dirs, _get_cwd
 from xonsh.environ import locate_binary
 from xonsh.foreign_shells import foreign_shell_data
 from xonsh.jobs import jobs, fg, bg, clean_jobs
-from xonsh.history import history_main
 from xonsh.platform import (ON_ANACONDA, ON_DARWIN, ON_WINDOWS, ON_FREEBSD,
-                            ON_NETBSD, scandir)
+                            ON_NETBSD)
 from xonsh.proc import foreground
 from xonsh.replay import replay_main
 from xonsh.timings import timeit_alias
-from xonsh.tools import (XonshError, argvquote, escape_windows_cmd_string,
-                         to_bool)
+from xonsh.tools import argvquote, escape_windows_cmd_string, to_bool
 from xonsh.xontribs import xontribs_main
 
 import xonsh.completers._aliases as xca
+import xonsh.history.main as xhm
 import xonsh.xoreutils.which as xxw
 
 
@@ -322,7 +321,6 @@ def xexec(args, stdin=None):
                           '\n'.format(e.args[1], args[0]), 1)
 
 
-
 class AWitchAWitch(argparse.Action):
     SUPPRESS = '==SUPPRESS=='
 
@@ -393,7 +391,7 @@ def make_default_aliases():
         'source-bash':  ['source-foreign', 'bash', '--sourcer=source'],
         'source-cmd': source_cmd,
         'source-foreign': source_foreign,
-        'history': history_main,
+        'history': xhm.history_main,
         'replay': replay_main,
         'trace': trace,
         'timeit': timeit_alias,
@@ -401,7 +399,7 @@ def make_default_aliases():
         'scp-resume': ['rsync', '--partial', '-h', '--progress', '--rsh=ssh'],
         'showcmd': showcmd,
         'ipynb': ['jupyter', 'notebook', '--no-browser'],
-        'which': xxw.which_main,
+        'which': xxw.which,
         'xontrib': xontribs_main,
         'completer': xca.completer_alias
     }
