@@ -4,7 +4,7 @@
 * env
 * gitstatus
 * job
-* vc_branch
+* vc
 * base
 
 """
@@ -181,6 +181,7 @@ GitStatus = collections.namedtuple('GitStatus',
                                     'untracked', 'changed', 'conflicts',
                                     'staged', 'stashed', 'operations'])
 
+
 def _check_output(*args, **kwargs):
     kwargs.update(dict(env=builtins.__xonsh_env__.detype(),
                        stderr=subprocess.DEVNULL,
@@ -286,8 +287,8 @@ def gitstatus():
     operations = _gitoperation(gitdir)
 
     return GitStatus(branch, num_ahead, num_behind,
-            untracked, changed, conflicts, staged, stashed,
-            operations)
+                     untracked, changed, conflicts, staged, stashed,
+                     operations)
 
 
 def gitstatus_prompt():
@@ -339,31 +340,41 @@ def _current_job():
             return s
 
 #
-# vc_branch
+# vc
 #
 # -*- coding: utf-8 -*-
 """Prompt formatter for simple version control branchs"""
+# pylint:disable=no-member, invalid-name
 
-# amalgamated builtins
 # amalgamated os
-# amalgamated subprocess
 sys = _LazyModule.load('sys', 'sys')
-threading = _LazyModule.load('threading', 'threading')
 queue = _LazyModule.load('queue', 'queue')
-time = _LazyModule.load('time', 'time')
-warnings = _LazyModule.load('warnings', 'warnings')
-# amalgamated xonsh.platform
+# amalgamated builtins
+threading = _LazyModule.load('threading', 'threading')
+# amalgamated subprocess
 # amalgamated xonsh.tools
 def _get_git_branch(q):
     try:
-        status = subprocess.check_output(['git', 'status'],
-                                         stderr=subprocess.DEVNULL)
-    except (subprocess.CalledProcessError, OSError):
+        branches = xt.decode_bytes(subprocess.check_output(
+            ['git', 'branch'],
+            stderr=subprocess.DEVNULL
+        )).splitlines()
+    except (subprocess.CalledProcessError, OSError, FileNotFoundError):
         q.put(None)
     else:
-        info = xt.decode_bytes(status)
-        branch = info.splitlines()[0].split()[-1]
-        q.put(branch)
+        for branch in branches:
+            if not branch.startswith('* '):
+                continue
+            elif branch.endswith(')'):
+                branch = branch.split()[-1][:-1]
+            else:
+                branch = branch.split()[-1]
+
+            q.put(branch)
+            break
+        else:
+            q.put(None)
+
 
 def get_git_branch():
     """Attempts to find the current git branch. If this could not
@@ -383,43 +394,42 @@ def get_git_branch():
     return branch
 
 
-def _get_parent_dir_for(path, dir_name, timeout):
-    # walk up the directory tree to see if we are inside an hg repo
-    # the timeout makes sure that we don't thrash the file system
-    previous_path = ''
-    t0 = time.time()
-    while path != previous_path and ((time.time() - t0) < timeout):
-        if os.path.isdir(os.path.join(path, dir_name)):
-            return path
-        previous_path = path
-        path, _ = os.path.split(path)
-    return (path == previous_path)
-
-
-def get_hg_branch(cwd=None, root=None):
+def get_hg_branch(root=None):
+    """Try to get the mercurial branch of the current directory,
+    return None if not in a repo or subprocess.TimeoutExpired if timed out.
+    """
     env = builtins.__xonsh_env__
-    cwd = env['PWD']
-    root = _get_parent_dir_for(cwd, '.hg', env['VC_BRANCH_TIMEOUT'])
-    if not isinstance(root, str):
-        # Bail if we are not in a repo or we timed out
-        if root:
-            return None
-        else:
-            return subprocess.TimeoutExpired(['hg'], env['VC_BRANCH_TIMEOUT'])
-    # get branch name
-    branch_path = os.path.sep.join([root, '.hg', 'branch'])
-    if os.path.exists(branch_path):
-        with open(branch_path, 'r') as branch_file:
-            branch = branch_file.read()
+    timeout = env['VC_BRANCH_TIMEOUT']
+    try:
+        root = subprocess.check_output(['hg', 'root'], timeout=timeout,
+                                       stderr=subprocess.DEVNULL)
+    except subprocess.TimeoutExpired:
+        return subprocess.TimeoutExpired(['hg'], timeout)
+    except (subprocess.CalledProcessError, FileNotFoundError):
+        # not in repo or command not in PATH
+        return None
     else:
-        branch = 'default'
+        root = xt.decode_bytes(root).strip()
+    if env.get('VC_HG_SHOW_BRANCH'):
+        # get branch name
+        branch_path = os.path.sep.join([root, '.hg', 'branch'])
+        if os.path.exists(branch_path):
+            with open(branch_path, 'r') as branch_file:
+                branch = branch_file.read()
+        else:
+            branch = 'default'
+    else:
+        branch = ''
     # add bookmark, if we can
     bookmark_path = os.path.sep.join([root, '.hg', 'bookmarks.current'])
     if os.path.exists(bookmark_path):
         with open(bookmark_path, 'r') as bookmark_file:
             active_bookmark = bookmark_file.read()
-        branch = "{0}, {1}".format(*(b.strip(os.linesep) for b in
-                                     (branch, active_bookmark)))
+        if env.get('VC_HG_SHOW_BRANCH') is True:
+            branch = "{0}, {1}".format(*(b.strip(os.linesep) for b in
+                                         (branch, active_bookmark)))
+        else:
+            branch = active_bookmark.strip(os.linesep)
     else:
         branch = branch.strip(os.linesep)
     return branch
@@ -436,7 +446,7 @@ def _first_branch_timeout_message():
     _FIRST_BRANCH_TIMEOUT = False
     print('xonsh: branch timeout: computing the branch name, color, or both '
           'timed out while formatting the prompt. You may avoid this by '
-          'increaing the value of $VC_BRANCH_TIMEOUT or by removing branch '
+          'increasing the value of $VC_BRANCH_TIMEOUT or by removing branch '
           'fields, like {curr_branch}, from your $PROMPT. See the FAQ '
           'for more details. This message will be suppressed for the remainder '
           'of this session. To suppress this message permanently, set '
@@ -444,20 +454,24 @@ def _first_branch_timeout_message():
           file=sys.stderr)
 
 
-def current_branch(pad=NotImplemented):
+def current_branch():
     """Gets the branch for a current working directory. Returns an empty string
     if the cwd is not a repository.  This currently only works for git and hg
     and should be extended in the future.  If a timeout occurred, the string
     '<branch-timeout>' is returned.
     """
-    if pad is not NotImplemented:
-        warnings.warn("The pad argument of current_branch has no effect now "
-                      "and will be removed in the future")
     branch = None
     cmds = builtins.__xonsh_commands_cache__
-    if cmds.lazy_locate_binary('git') or cmds.is_empty():
+    # check for binary only once
+    if cmds.is_empty():
+        has_git = bool(cmds.locate_binary('git'))
+        has_hg = bool(cmds.locate_binary('hg'))
+    else:
+        has_git = bool(cmds.lazy_locate_binary('git'))
+        has_hg = bool(cmds.lazy_locate_binary('hg'))
+    if has_git:
         branch = get_git_branch()
-    if (cmds.lazy_locate_binary('hg') or cmds.is_empty()) and not branch:
+    if not branch and has_hg:
         branch = get_hg_branch()
     if isinstance(branch, subprocess.TimeoutExpired):
         branch = '<branch-timeout>'
@@ -474,7 +488,7 @@ def _git_dirty_working_directory(q, include_untracked):
         else:
             cmd.append('--untracked-files=no')
         status = subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
-    except (subprocess.CalledProcessError, OSError):
+    except (subprocess.CalledProcessError, OSError, FileNotFoundError):
         q.put(None)
     if status is not None:
         return q.put(bool(status))
@@ -498,7 +512,7 @@ def git_dirty_working_directory(include_untracked=False):
 
 def hg_dirty_working_directory():
     """Computes whether or not the mercurial working directory is dirty or not.
-    If this cannot be deterimined, None is returned.
+    If this cannot be determined, None is returned.
     """
     env = builtins.__xonsh_env__
     cwd = env['PWD']
@@ -508,24 +522,25 @@ def hg_dirty_working_directory():
     denv['HGRCPATH'] = ''
     try:
         s = subprocess.check_output(['hg', 'identify', '--id'],
-                                    stderr=subprocess.PIPE, cwd=cwd, timeout=vcbt,
-                                    universal_newlines=True, env=denv)
+                                    stderr=subprocess.PIPE, cwd=cwd,
+                                    timeout=vcbt, universal_newlines=True,
+                                    env=denv)
         return s.strip(os.linesep).endswith('+')
     except (subprocess.CalledProcessError, subprocess.TimeoutExpired,
             FileNotFoundError):
         return None
 
 
-def dirty_working_directory(cwd=None):
+def dirty_working_directory():
     """Returns a boolean as to whether there are uncommitted files in version
     control repository we are inside. If this cannot be determined, returns
     None. Currently supports git and hg.
     """
     dwd = None
     cmds = builtins.__xonsh_commands_cache__
-    if cmds.lazy_locate_binary('git') or cmds.is_empty():
+    if cmds.lazy_locate_binary('git'):
         dwd = git_dirty_working_directory()
-    if (cmds.lazy_locate_binary('hg') or cmds.is_empty()) and (dwd is None):
+    if cmds.lazy_locate_binary('hg') and dwd is None:
         dwd = hg_dirty_working_directory()
     return dwd
 
@@ -562,7 +577,7 @@ def branch_bg_color():
 # base
 #
 # -*- coding: utf-8 -*-
-"""Base prompt, provides FORMATTER_DICT and prompt related functions"""
+"""Base prompt, provides PROMPT_FIELDS and prompt related functions"""
 
 # amalgamated builtins
 itertools = _LazyModule.load('itertools', 'itertools')
@@ -577,10 +592,75 @@ string = _LazyModule.load('string', 'string')
 # amalgamated xonsh.prompt.cwd
 # amalgamated xonsh.prompt.job
 # amalgamated xonsh.prompt.env
-# amalgamated xonsh.prompt.vc_branch
+# amalgamated xonsh.prompt.vc
 # amalgamated xonsh.prompt.gitstatus
+@xt.lazyobject
+def DEFAULT_PROMPT():
+    return default_prompt()
+
+
+class PromptFormatter:
+    """Class that holds all the related prompt formatting methods,
+    uses the ``PROMPT_FIELDS`` envvar (no color formatting).
+    """
+
+    def __init__(self):
+        self.cache = {}
+
+    def __call__(self, template=DEFAULT_PROMPT, fields=None):
+        """Formats a xonsh prompt template string."""
+        if fields is None:
+            self.fields = builtins.__xonsh_env__.get('PROMPT_FIELDS', PROMPT_FIELDS)
+        else:
+            self.fields = fields
+        try:
+            prompt = self._format_prompt(template=template)
+        except Exception:
+            return _failover_template_format(template)
+        # keep cache only during building prompt
+        self.cache.clear()
+        return prompt
+
+    def _format_prompt(self, template=DEFAULT_PROMPT):
+        template = template() if callable(template) else template
+        toks = []
+        for literal, field, spec, conv in _FORMATTER.parse(template):
+            toks.append(literal)
+            entry = self._format_field(field, spec, conv)
+            if entry is not None:
+                toks.append(entry)
+        return ''.join(toks)
+
+    def _format_field(self, field, spec, conv):
+        if field is None:
+            return
+        elif field.startswith('$'):
+            val = builtins.__xonsh_env__[field[1:]]
+            return _format_value(val, spec, conv)
+        elif field in self.fields:
+            val = self._get_field_value(field)
+            return _format_value(val, spec, conv)
+        else:
+            # color or unkown field, return as is
+            return '{' + field + '}'
+
+    def _get_field_value(self, field):
+        field_value = self.fields[field]
+        if field_value in self.cache:
+            return self.cache[field_value]
+        try:
+            value = field_value() if callable(field_value) else field_value
+            self.cache[field_value] = value
+        except Exception:
+            print('prompt: error: on field {!r}'
+                  ''.format(field), file=sys.stderr)
+            xt.print_exception()
+            value = '(ERROR:{})'.format(field)
+        return value
+
+
 @xl.lazyobject
-def FORMATTER_DICT():
+def PROMPT_FIELDS():
     return dict(
         user=os.environ.get('USERNAME' if xp.ON_WINDOWS else 'USER', '<user>'),
         prompt_end='#' if xt.is_superuser() else '$',
@@ -622,19 +702,6 @@ def default_prompt():
     return dp
 
 
-@xt.lazyobject
-def DEFAULT_PROMPT():
-    return default_prompt()
-
-
-def _get_fmtter(formatter_dict=None):
-    if formatter_dict is None:
-        fmtter = builtins.__xonsh_env__.get('FORMATTER_DICT', FORMATTER_DICT)
-    else:
-        fmtter = formatter_dict
-    return fmtter
-
-
 def _failover_template_format(template):
     if callable(template):
         try:
@@ -647,56 +714,6 @@ def _failover_template_format(template):
     return template
 
 
-def partial_format_prompt(template=DEFAULT_PROMPT, formatter_dict=None):
-    """Formats a xonsh prompt template string."""
-    try:
-        return _partial_format_prompt_main(template=template,
-                                           formatter_dict=formatter_dict)
-    except Exception:
-        return _failover_template_format(template)
-
-
-def _partial_format_prompt_main(template=DEFAULT_PROMPT, formatter_dict=None):
-    template = template() if callable(template) else template
-    fmtter = _get_fmtter(formatter_dict)
-    bopen = '{'
-    bclose = '}'
-    colon = ':'
-    expl = '!'
-    toks = []
-    for literal, field, spec, conv in _FORMATTER.parse(template):
-        toks.append(literal)
-        if field is None:
-            continue
-        elif field.startswith('$'):
-            val = builtins.__xonsh_env__[field[1:]]
-            val = _format_value(val, spec, conv)
-            toks.append(val)
-        elif field in fmtter:
-            v = fmtter[field]
-            try:
-                val = v() if callable(v) else v
-            except Exception as err:
-                print('prompt: error: on field {!r}'
-                      ''.format(field), file=sys.stderr)
-                xt.print_exception()
-                toks.append('(ERROR:{})'.format(field))
-                continue
-            val = _format_value(val, spec, conv)
-            toks.append(val)
-        else:
-            toks.append(bopen)
-            toks.append(field)
-            if conv is not None and len(conv) > 0:
-                toks.append(expl)
-                toks.append(conv)
-            if spec is not None and len(spec) > 0:
-                toks.append(colon)
-                toks.append(spec)
-            toks.append(bclose)
-    return ''.join(toks)
-
-
 @xt.lazyobject
 def RE_HIDDEN():
     return re.compile('\001.*?\002')
@@ -749,7 +766,7 @@ def multiline_prompt(curr=''):
     return rtn
 
 
-def is_template_string(template, formatter_dict=None):
+def is_template_string(template, PROMPT_FIELDS=None):
     """Returns whether or not the string is a valid template."""
     template = template() if callable(template) else template
     try:
@@ -757,10 +774,10 @@ def is_template_string(template, formatter_dict=None):
     except ValueError:
         return False
     included_names.discard(None)
-    if formatter_dict is None:
-        fmtter = builtins.__xonsh_env__.get('FORMATTER_DICT', FORMATTER_DICT)
+    if PROMPT_FIELDS is None:
+        fmtter = builtins.__xonsh_env__.get('PROMPT_FIELDS', PROMPT_FIELDS)
     else:
-        fmtter = formatter_dict
+        fmtter = PROMPT_FIELDS
     known_names = set(fmtter.keys())
     return included_names <= known_names
 
